// StaticTerm.cpp : implementation file
//

#include "stdafx.h"
#include "ilvideo80.h"
#include "StaticTerm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define	ROP_DSPDxax	0x00E20746

// ROMAN8 Unicode table
const WCHAR CStaticTerm::wcRoman8[] =
{
	0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020,
	0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020,
	0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020,
	0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020,
	0x0020, 0x00C0, 0x00C2, 0x00C8, 0x00CA, 0x00CB, 0x00CE, 0x00CF,
	0x00B4, 0x0060, 0x02C6, 0x00A8, 0x02DC, 0x00D9, 0x00FB, 0x20A4,
	0x00AF, 0x0020, 0x0020, 0x00B0, 0x00C7, 0x00E7, 0x00D1, 0x00F1,
	0x00A1, 0x00BF, 0x00A4, 0x00A3, 0x00A5, 0x00A7, 0x0192, 0x00A2,
	0x00E2, 0x00EA, 0x00F4, 0x00FB, 0x00E1, 0x00E9, 0x00F3, 0x00FA,
	0x00E0, 0x00E8, 0x00F2, 0x00F9, 0x00E4, 0x00EB, 0x00F6, 0x00FC,
	0x00C5, 0x00EE, 0x00D8, 0x00C6, 0x00E5, 0x00ED, 0x00F8, 0x00E6,
	0x00C4, 0x00EC, 0x00D6, 0x00DC, 0x00C9, 0x00EF, 0x00DF, 0x00D4,
	0x00C1, 0x00C3, 0x00E3, 0x00D0, 0x00F0, 0x00CD, 0x00CC, 0x00D3,
	0x00D2, 0x00D5, 0x00F5, 0x0160, 0x0161, 0x00DA, 0x0178, 0x00FF,
	0x00DE, 0x00FE, 0x0020, 0x0020, 0x0020, 0x0020, 0x2014, 0x00BC,
	0x00BD, 0x00AA, 0x00BA, 0x00AB, 0x25A0, 0x00BB, 0x00B1, 0x0020
};

/////////////////////////////////////////////////////////////////////////////
// CStaticTerm

IMPLEMENT_DYNAMIC(CStaticTerm, CStatic)

CStaticTerm::CStaticTerm(INT nCols,INT nRows,DWORD dwRefresh)
	: m_dwRefresh(dwRefresh)				// output refresh time in ms
{
	BITMAP CurBitMap;

	InitializeCriticalSection(&m_csLock);

	m_rgbTextColor = RGB(0xFF,0xFF,0xFF);	// white text color
	m_rgbBkColor   = RGB(0x00,0x00,0x00);	// black background color

	m_nBlinkFreq = 257;						// US blink frequency in ms

	m_nCols = nCols;						// no. of visible columns
	m_nRows = nRows;						// no. of visible rows

	m_nWidth = 0;							// character width
	m_nHeight = 0;							// character height

	m_xCurPos = 0; 							// x cursor position
	m_yCurPos = 0; 							// y cursor position

	m_hEvent = NULL;
	m_hWorkerThread = NULL;
	m_bCursorAct = false;					// showing text
	m_eCurState = CUR_OFF;					// cursor off

	m_bASCII = true;						// ASCII character set

	// init monitor character bitmap
	VERIFY(m_BmpMonChars.LoadBitmap((nCols >= 80) ? IDB_MONCHRS : IDB_MONCHRL));

	m_BmpMonChars.GetBitmap(&CurBitMap);
	m_nMonChrWidth = CurBitMap.bmWidth / 31;
	m_nMonChrHeight = CurBitMap.bmHeight;

	m_strBuffer = _T("");					// intermediate buffer
	m_hDrawThread = NULL;
	m_bRunning = false;
	m_hDrawEvent = NULL;					// event to start drawing timer

	StartDrawWorker();						// start worker thread for buffered output
}

CStaticTerm::~CStaticTerm()
{
	m_BmpMonChars.DeleteObject();
	m_Font80.DeleteObject();
	m_Font40.DeleteObject();

	DeleteCriticalSection(&m_csLock);
}

BEGIN_MESSAGE_MAP(CStaticTerm, CStatic)
	//{{AFX_MSG_MAP(CStaticTerm)
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CStaticTerm message handlers

#if _MSC_VER <= 1200 // MSVC6 and earlier
BOOL CStaticTerm::OnChildNotify(UINT message, WPARAM wParam, LPARAM lParam, LRESULT* pLResult) 
{
	// TODO: Add your specialized code here and/or call the base class
	if (message == WM_DRAWITEM)
	{
		ASSERT(pLResult == NULL);			// no return value expected
		UNUSED(pLResult);					// unused in release builds
		DrawItem(reinterpret_cast<LPDRAWITEMSTRUCT>(lParam));
		return TRUE;
	}
	return CStatic::OnChildNotify(message, wParam, lParam, pLResult);
}
#endif

void CStaticTerm::PreSubclassWindow()
{
	// TODO: Add your specialized code here and/or call the base class
	ModifyStyle(0, SS_OWNERDRAW);

	LOGFONT lf;
	// create non antialiased fonts to disable ClearType
	GetFont()->GetLogFont(&lf);
	lf.lfQuality = NONANTIALIASED_QUALITY;

	CClientDC dc(this);

	lf.lfHeight = 8;
	lf.lfWidth = 0;

	lf.lfHeight = -MulDiv(lf.lfHeight,dc.GetDeviceCaps(LOGPIXELSY),72);
	m_Font80.CreateFontIndirect(&lf);

	lf.lfHeight = 11;

	lf.lfHeight = -MulDiv(lf.lfHeight,dc.GetDeviceCaps(LOGPIXELSY),72);
	lf.lfWidth = (-lf.lfHeight + 1) * 14 / 17;
	m_Font40.CreateFontIndirect(&lf);

	SetFont((m_nCols >= 80) ? &m_Font80 : &m_Font40,FALSE);

	CStatic::PreSubclassWindow();
}

void CStaticTerm::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	// TODO: Add your message handler code here and/or call default
	CBitmap bmpText,*pbmpText;
	CBitmap bmpMask,*pbmpMask;
	CDC dcMask,dcText;
	INT nLinCurPos,nLen;
	bool bCurInv;
	TEXTMETRIC tm;

	// draw bit not set
	if ((lpDrawItemStruct->itemAction & ODA_DRAWENTIRE) == 0) return;

	CRect rc = lpDrawItemStruct->rcItem;
	CDC *pDC = CDC::FromHandle(lpDrawItemStruct->hDC);

	pDC->GetTextMetrics(&tm);				// get character dimensions
	m_nWidth = tm.tmAveCharWidth;
	m_nHeight = tm.tmHeight;

	// create text DC
	dcText.CreateCompatibleDC(pDC);
	VERIFY(bmpText.CreateCompatibleBitmap(pDC,rc.Width(),rc.Height()));
	pbmpText = dcText.SelectObject(&bmpText);

	// create mask DC
	dcMask.CreateCompatibleDC(pDC);
	VERIFY(bmpMask.CreateCompatibleBitmap(pDC,rc.Width(),rc.Height()));
	pbmpMask = dcMask.SelectObject(&bmpMask);

	// fill mask DC
	VERIFY(dcMask.PatBlt(0,0,rc.Width(),rc.Height(),BLACKNESS));

	nLinCurPos = (m_yCurPos * (m_nCols + 2)) + m_xCurPos;
	bCurInv = false;						// draw cursor not inverse

	nLen = GetWindowTextLength();			// get length of window content

	// buffer for the text content
	LPTSTR pcBuffer = static_cast<LPTSTR>(new TCHAR[nLen+1]);
	if (pcBuffer != NULL)
	{
		GetWindowText(pcBuffer,nLen+1);		// get item text

		for (INT i = 0; i < nLen; ++i)
		{
			// skip "\r\n" at end of each column because they are not part of the data
			if (   i % (m_nCols + 2) == (m_nCols + 0) 
				|| i % (m_nCols + 2) == (m_nCols + 1))
			{
				ASSERT(   (pcBuffer[i] == _T('\r') && i % (m_nCols + 2) == (m_nCols + 0))
					   || (pcBuffer[i] == _T('\n') && i % (m_nCols + 2) == (m_nCols + 1)));
				continue;
			}

			if ((pcBuffer[i] & 0x80) != 0)
			{
				if (m_bASCII)				// ASCII character set
				{
					if (i == nLinCurPos)	// linear address of cursor
					{
						bCurInv = true;		// draw cursor inverse
					}

					// modify pcBuffer -> if Bit 7 is set,
					// clear it and set the invert mask at this character
					if (pcBuffer[i] < 0xFF)
					{
						pcBuffer[i] &= ~0x80; // clear bit

						// set inverse mask
						VERIFY(dcMask.PatBlt((i % (m_nCols + 2)) * m_nWidth,
											 (i / (m_nCols + 2)) * m_nHeight,
											 m_nWidth,
											 m_nHeight,
											 WHITENESS));
					}
					else					// special character 255 drawing
					{
						// white invers
						pcBuffer[i] = _T(' ');
					}
				}
				else						// Roman-8 character set
				{	
					// translate Roman-8 to windows font
					pcBuffer[i] = static_cast<TCHAR>(wcRoman8[static_cast<BYTE>(pcBuffer[i]) - 0x80]);
				}
			}
			else if (pcBuffer[i] == 0x7F)	// special character 127 drawing
			{
				// set inverse mask
				VERIFY(dcMask.PatBlt((i % (m_nCols + 2)) * m_nWidth,
									 (i / (m_nCols + 2)) * m_nHeight,
									 m_nWidth,
									 m_nHeight,
									 WHITENESS));

				pcBuffer[i] = _T(' ');
			}

			// non printable characters 0-31 and 128-159 (mapped to 0-31 in ASCII mode)
			if (static_cast<WORD>(pcBuffer[i]) < _T(' '))
			{
				CDC     dcChars;
				CBitmap *pbmpChars;

				ASSERT(   static_cast<WORD>(pcBuffer[i]) >= 0
					   && static_cast<WORD>(pcBuffer[i]) < _T(' '));

				// draw special characters from bitmap
				dcChars.CreateCompatibleDC(pDC);
				pbmpChars = dcChars.SelectObject(&m_BmpMonChars);
				VERIFY(dcMask.StretchBlt((i % (m_nCols + 2)) * m_nWidth,
										 (i / (m_nCols + 2)) * m_nHeight,
										 m_nWidth,
										 m_nHeight,
										 &dcChars,
										 static_cast<WORD>(pcBuffer[i]) * m_nMonChrWidth,
										 0,
										 m_nMonChrWidth,
										 m_nMonChrHeight,
										 SRCINVERT));
				dcChars.SelectObject(pbmpChars);

				pcBuffer[i] = _T(' ');
			}
		}

		// fill text DC with B&W text
		dcText.SetTextColor(RGB(255,255,255));
		dcText.SetBkColor(RGB(0,0,0));

		CFont *pFontPrev = dcText.SelectObject(GetFont());
		dcText.DrawText(pcBuffer,nLen,&rc,DT_LEFT | DT_NOPREFIX | DT_NOCLIP);
		dcText.SelectObject(pFontPrev);

		delete[] pcBuffer;
	}

	// add text to the background invert mask
	VERIFY(dcMask.BitBlt(rc.left,
						 rc.top,
						 rc.Width(),
						 rc.Height(),
						 &dcText,
						 rc.left,
						 rc.top,
						 SRCINVERT));

	if (m_bCursorAct)						// show cursor
	{
		// select the correct cursor
		switch (m_eCurState)
		{
		case CUR_OFF:
			break;
		case CUR_REPLACE:
			VERIFY(dcMask.PatBlt(m_xCurPos * m_nWidth,
								 m_yCurPos * m_nHeight,
								 m_nWidth,
								 m_nHeight,
								 (bCurInv) ? BLACKNESS : WHITENESS));
			break;
		case CUR_INSERT:
			VERIFY(dcMask.PatBlt(m_xCurPos * m_nWidth,
								 m_yCurPos * m_nHeight,
								 m_nWidth,
								 m_nHeight,
								 (bCurInv) ? WHITENESS : BLACKNESS));
			VERIFY(dcMask.PatBlt(m_xCurPos * m_nWidth,
								 (m_yCurPos + 1) * m_nHeight - (tm.tmDescent + 1) / 2,
								 m_nWidth,
								 (tm.tmDescent + 1) / 2,
								 (bCurInv) ? BLACKNESS : WHITENESS));
			break;
		default: ASSERT(false);
		}
	}

	// convert the BW to a two color bitmap
	dcText.FillSolidRect(rc,m_rgbBkColor);	// fill with background color

	CBrush Brush,*pBrush;
	Brush.CreateSolidBrush(m_rgbTextColor);	// foreground color brush
	pBrush = dcText.SelectObject(&Brush);
	VERIFY(dcText.BitBlt(rc.left,
						 rc.top,
						 rc.Width(),
						 rc.Height(),
						 &dcMask,
						 rc.left,
						 rc.top,
						 ROP_DSPDxax));
	dcText.SelectObject(pBrush);
	VERIFY(Brush.DeleteObject());

	// update display
	VERIFY(pDC->BitBlt(rc.left,
					   rc.top,
					   rc.Width(),
					   rc.Height(),
					   &dcText,
					   rc.left,
					   rc.top,
					   SRCCOPY));

	dcMask.SelectObject(pbmpMask);			// delete mask
	VERIFY(bmpMask.DeleteObject());

	dcText.SelectObject(pbmpText);			// delete text
	VERIFY(bmpText.DeleteObject());
}

void CStaticTerm::OnDestroy()
{
	CStatic::OnDestroy();

	// TODO: Add your message handler code here
	StopBlinkingThread();					// destroy cursor thread
	StopDrawWorker();						// stop the worker thread
}

VOID CStaticTerm::InvalidateCursor()
{
	RECT rc;

	rc.left   = m_xCurPos * m_nWidth;
	rc.top    = m_yCurPos * m_nHeight;
	rc.right  = rc.left + m_nWidth;
	rc.bottom = rc.top  + m_nHeight;
	InvalidateRect(&rc,FALSE);
	return;
}

UINT CStaticTerm::ThreadCusorBlink(LPVOID pParam)
{
	CStaticTerm *p = reinterpret_cast<CStaticTerm *>(pParam);
	ASSERT(p != NULL && p->IsKindOf(RUNTIME_CLASS(CStaticTerm)));

	// show cursor
	p->m_bCursorAct = true;					// show cursor
	p->InvalidateCursor();

	while (WaitForSingleObject(p->m_hEvent,p->m_nBlinkFreq) == WAIT_TIMEOUT)
	{
		p->m_bCursorAct = !p->m_bCursorAct;	// toogle cursor / normal char
		p->InvalidateCursor();
	}

	// force to exit with normal char
	p->m_bCursorAct = false;				// show text
	p->InvalidateCursor();
	return 0;
}

//
// start cursor blinking thread
//
bool CStaticTerm::StartBlinkingThread()
{
	DWORD dwThreadID;

	bool bSucc = false;

	if (m_hEvent == NULL && m_hWorkerThread == NULL)
	{
		// create shutdown event
		VERIFY(m_hEvent = CreateEvent(NULL,FALSE,FALSE,NULL));

		// create blinking thread
		m_hWorkerThread = CreateThread(NULL,
									   0,
									   (LPTHREAD_START_ROUTINE) ThreadCusorBlink,
									   (LPVOID) this,
									   0,
									   &dwThreadID);
		ASSERT(m_hWorkerThread);
		bSucc = (m_hWorkerThread != NULL);
	}
	return bSucc;
}

//
// stop cursor blinking thread
//
VOID CStaticTerm::StopBlinkingThread()
{
	if (m_hWorkerThread)					// shut down thread
	{
		SetEvent(m_hEvent); 				// set finish event

		// wait for thread down
		WaitForSingleObject(m_hWorkerThread,INFINITE);
		CloseHandle(m_hWorkerThread);
		m_hWorkerThread = NULL;
	}

	if (m_hEvent)							// close event
	{
		CloseHandle(m_hEvent);
		m_hEvent = NULL;
	}
	return;
}

//
// get translated Roman8 character
//
WCHAR CStaticTerm::GetRoman8Char(BYTE byAsciiChar)
{
	WCHAR wc = byAsciiChar;
	if (byAsciiChar >= 0x80)
	{
		ASSERT(byAsciiChar - 0x80 <= ARRAYSIZEOF(wcRoman8));
		wc = wcRoman8[byAsciiChar - 0x80];
	}
	return wc;
}

//
// set new text and background color
//
VOID CStaticTerm::SetColors(COLORREF rgbTextColor, COLORREF rgbBkColor)
{
	m_rgbTextColor = rgbTextColor;			// text color
	m_rgbBkColor = rgbBkColor;				// background color
	Invalidate(FALSE);						// redraw all
	return;
}

//
// switch between ASCII and Roman-8 character set
//
VOID CStaticTerm::SetAsciiMode(bool bASCII)
{
	m_bASCII = bASCII;
	Invalidate(FALSE);						// redraw all
	return;
}

//
// set new size of terminal window
//
VOID CStaticTerm::SetSize(INT nCols,INT nRows)
{
	TEXTMETRIC tm;

	SetBufferText(_T(""));					// clear output window

	m_nCols = nCols;						// new size
	m_nRows = nRows;

	// init monitor character bitmap
	m_BmpMonChars.DeleteObject();
	VERIFY(m_BmpMonChars.LoadBitmap((m_nCols >= 80) ? IDB_MONCHRS : IDB_MONCHRL));

	BITMAP CurBitMap;
	m_BmpMonChars.GetBitmap(&CurBitMap);
	m_nMonChrWidth = CurBitMap.bmWidth / 31;
	m_nMonChrHeight = CurBitMap.bmHeight;

	// change font
	SetFont((m_nCols >= 80) ? &m_Font80 : &m_Font40,FALSE);

	CClientDC dc(this);
	CFont* pFont = dc.SelectObject(GetFont());
	dc.GetTextMetrics(&tm);					// get character dimensions
	dc.SelectObject(pFont);

	// adjust text window to middle of tube monitor background rectangle
	CRect rectBorder;

	GetParent()->GetDlgItem(IDC_BORDER)->GetWindowRect(&rectBorder);

	LONG lWidth  = tm.tmAveCharWidth * m_nCols;
	LONG lHeight = tm.tmHeight * m_nRows;

	GetParent()->ScreenToClient(&rectBorder);
	rectBorder.left += (rectBorder.Width()  - lWidth)  / 2;
	rectBorder.top  += (rectBorder.Height() - lHeight) / 2;

	// resize and repositioning window
	SetWindowPos(NULL,rectBorder.left,rectBorder.top,lWidth,lHeight,SWP_NOZORDER);
	return;
}

//
// set new cursor position and state
//
VOID CStaticTerm::SetCursor(INT x, INT y, enum CUR_STATE eState)
{
	ASSERT(m_eCurState >= CUR_OFF && m_eCurState <= CUR_INSERT);
	ASSERT(eState      >= CUR_OFF && eState      <= CUR_INSERT);

	x %= m_nCols;							// limit coordinates to range
	y %= m_nRows;

	// cursor state or position change
	if (eState != m_eCurState || x != m_xCurPos || y != m_yCurPos)
	{
		m_eCurState = CUR_OFF;				// switch cursor off
		StopBlinkingThread();
		InvalidateCursor();					// redraw character under cursor
	}

	m_xCurPos = x;							// save new cursor position
	m_yCurPos = y;

	if (eState != CUR_OFF)					// switch on cursor
	{
		StartBlinkingThread();
	}
	m_eCurState = eState;					// save new state
	return;
}

//
// worker thread for drawing
//
VOID CStaticTerm::SetBufferText(LPCTSTR lpszText)
{
	EnterCriticalSection(&m_csLock);
	{
		BOOL bTrigger = m_strBuffer.IsEmpty();

		m_strBuffer = (LPCTSTR) lpszText;
		if (bTrigger)
		{
			SetEvent(m_hDrawEvent);			// start update timer
		}
	}
	LeaveCriticalSection(&m_csLock);
	return;
}

//
// flush buffered text
//
VOID CStaticTerm::FlushText()
{
	EnterCriticalSection(&m_csLock);
	{
		if (!m_strBuffer.IsEmpty())
		{
			// write buffer
			SetWindowText(m_strBuffer);
			m_strBuffer.Empty();
		}
	}
	LeaveCriticalSection(&m_csLock);
	return;
}

//
// worker thread for drawing
//
UINT CStaticTerm::DrawThread(LPVOID pParam)
{
	CStaticTerm *p = reinterpret_cast<CStaticTerm *>(pParam);

	DWORD dwTimeout = INFINITE;

	ASSERT(p != NULL);

	while (p->m_bRunning)
	{
		switch (WaitForSingleObject(p->m_hDrawEvent,dwTimeout))
		{
		case WAIT_TIMEOUT: // update display output
			p->FlushText();
			dwTimeout = INFINITE;
			break;
		case WAIT_OBJECT_0: // trigger output
			dwTimeout = p->m_dwRefresh;
			break;
		}
	}
	return 0;
}

//
// start drawing worker thread
//
bool CStaticTerm::StartDrawWorker()
{
	DWORD dwThreadID;

	bool bSucc = false;

	m_bRunning = true;

	if (m_hDrawEvent == NULL && m_hDrawThread == NULL)
	{
		// create drawing timer event
		VERIFY(m_hDrawEvent = CreateEvent(NULL,FALSE,FALSE,NULL));

		// create drawing thread
		m_hDrawThread = CreateThread(NULL,
									 0,
									 (LPTHREAD_START_ROUTINE) DrawThread,
									 (LPVOID) this,
									 0,
									 &dwThreadID);
		ASSERT(m_hDrawThread);

		bSucc = (m_hDrawThread != NULL);
	}
	return bSucc;
}

//
// shut down drawing worker thread
//
VOID CStaticTerm::StopDrawWorker()
{
	if (m_hDrawThread)						// worker thread running
	{
		m_bRunning = false;					// exit worker thread
		SetEvent(m_hDrawEvent);				// kill waiting thread

		// wait for worker thread down
		WaitForSingleObject(m_hDrawThread,INFINITE);
		CloseHandle(m_hDrawThread);
		m_hDrawThread = NULL;

		FlushText();						// write rest of buffer
	}

	if (m_hDrawEvent)						// close event
	{
		CloseHandle(m_hDrawEvent);
		m_hDrawEvent = NULL;
	}
	return;
}
